from .otBase import BaseTTXConverter


class table__f_e_a_t(BaseTTXConverter):
    """Feature name table

    The feature name table is an AAT (Apple Advanced Typography) table for
    storing font features, settings, and their human-readable names. It should
    not be confused with the ``Feat`` table or the OpenType Layout ``GSUB``/``GPOS``
    tables.

    See also https://developer.apple.com/fonts/TrueType-Reference-Manual/RM06/Chap6feat.html
    """

    pass
