"""IFT - Incremental Font Transfer table."""

# Reference: https://github.com/googlefonts/fontations/blob/main/read-fonts/src/tables/ift.rs


from .otBase import BaseTTXConverter


class table_I_F_T_X_(BaseTTXConverter):
    """Incremental Font Transfer table."""

    pass
