# DON'T EDIT! This file is generated by MetaTools/buildTableList.py.
def _moduleFinderHint():
    """Dummy function to let modulefinder know what tables may be
    dynamically imported. Generated by MetaTools/buildTableList.py.

        >>> _moduleFinderHint()
    """
    from . import B_A_S_E_
    from . import C_B_D_T_
    from . import C_B_L_C_
    from . import C_F_F_
    from . import C_F_F__2
    from . import C_O_L_R_
    from . import C_P_A_L_
    from . import D_S_I_G_
    from . import D__e_b_g
    from . import E_B_D_T_
    from . import E_B_L_C_
    from . import F_F_T_M_
    from . import F__e_a_t
    from . import G_D_E_F_
    from . import G_P_O_S_
    from . import G_S_U_B_
    from . import G_V_A_R_
    from . import G__l_a_t
    from . import G__l_o_c
    from . import H_V_A_R_
    from . import I_F_T_
    from . import I_F_T_X_
    from . import J_S_T_F_
    from . import L_T_S_H_
    from . import M_A_T_H_
    from . import M_V_A_R_
    from . import O_S_2f_2
    from . import S_T_A_T_
    from . import S_V_G_
    from . import S__i_l_f
    from . import S__i_l_l
    from . import T_S_I_B_
    from . import T_S_I_C_
    from . import T_S_I_D_
    from . import T_S_I_J_
    from . import T_S_I_P_
    from . import T_S_I_S_
    from . import T_S_I_V_
    from . import T_S_I__0
    from . import T_S_I__1
    from . import T_S_I__2
    from . import T_S_I__3
    from . import T_S_I__5
    from . import T_T_F_A_
    from . import V_A_R_C_
    from . import V_D_M_X_
    from . import V_O_R_G_
    from . import V_V_A_R_
    from . import _a_n_k_r
    from . import _a_v_a_r
    from . import _b_g_c_l
    from . import _b_s_l_n
    from . import _c_i_d_g
    from . import _c_m_a_p
    from . import _c_v_a_r
    from . import _c_v_t
    from . import _f_e_a_t
    from . import _f_p_g_m
    from . import _f_v_a_r
    from . import _g_a_s_p
    from . import _g_c_i_d
    from . import _g_l_y_f
    from . import _g_v_a_r
    from . import _h_d_m_x
    from . import _h_e_a_d
    from . import _h_h_e_a
    from . import _h_m_t_x
    from . import _k_e_r_n
    from . import _l_c_a_r
    from . import _l_o_c_a
    from . import _l_t_a_g
    from . import _m_a_x_p
    from . import _m_e_t_a
    from . import _m_o_r_t
    from . import _m_o_r_x
    from . import _n_a_m_e
    from . import _o_p_b_d
    from . import _p_o_s_t
    from . import _p_r_e_p
    from . import _p_r_o_p
    from . import _s_b_i_x
    from . import _t_r_a_k
    from . import _v_h_e_a
    from . import _v_m_t_x


if __name__ == "__main__":
    import doctest, sys

    sys.exit(doctest.testmod().failed)
